from functools import lru_cache
from os.path import basename
from typing import Optional
import argparse
import logging
import sys


@lru_cache()
def setup_logging_stdout(name: Optional[str] = None,
                         verbose: bool = False) -> logging.Logger:
    name = name if name else basename(sys.argv[0])
    log = logging.getLogger(name)
    log.setLevel(logging.DEBUG if verbose else logging.INFO)
    channel = logging.StreamHandler(sys.stdout)
    channel.setFormatter(logging.Formatter('%(message)s'))
    channel.setLevel(logging.DEBUG if verbose else logging.INFO)
    log.addHandler(channel)
    return log


def common_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument('-u', '--username', required=False, help='Xirvik user')
    parser.add_argument('-p',
                        '--password',
                        required=False,
                        help='Xirvik password')
    parser.add_argument(
        '-r',
        '--max-retries',
        type=int,
        default=10,
        help='Number of retries for each request (passed to client)')
    parser.add_argument('-v',
                        '--verbose',
                        action='store_true',
                        help='Enable verbose logging')
    parser.add_argument(
        '--backoff-factor',
        type=int,
        default=5,
        help=('Back-off factor used when calculating time to wait to retry '
              'a failed request'))
    parser.add_argument('--netrc', required=False, help='netrc file path')
    parser.add_argument('host', nargs=1, help='Host name')
    return parser



"""SFTP client like paramiko's with extra features."""
from datetime import datetime
from math import ceil, floor
from os import chmod, makedirs, utime
from os.path import basename, dirname, isdir, join as path_join, realpath
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, cast
import inspect
import logging
import os
import socket

from humanize import naturaldelta, naturalsize
from paramiko import SFTPAttributes, SFTPClient as OriginalSFTPClient, SFTPFile
from paramiko.client import SSHClient
from paramiko.sftp import SFTPError

from .typing import Method0, Method1

__all__ = (
    'SFTPClient',
    'LOG_NAME',
)

LOG_NAME = 'xirvik.sftp'
LOG_INTERVAL = 60


class SFTPClient:
    """Dynamic extension on paramiko's SFTPClient."""
    chdir: Method1['SFTPClient', str, Optional[str]]
    getcwd: Method0['SFTPClient', str]
    listdir_iter: Callable[..., Iterator[Any]]

    MAX_PACKET_SIZE: int = SFTPFile.__dict__['MAX_REQUEST_SIZE']

    original_arguments: Dict[str, Any] = {}
    debug: bool = False

    _log = logging.getLogger(LOG_NAME)
    _dircache: List[str] = []

    def __init__(self, **kwargs: Any):
        """Constructor."""
        self.original_arguments = kwargs.copy()
        self._connect(**kwargs)

    def __enter__(self) -> 'SFTPClient':
        """For use with a with statement."""
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """For use with a with statement."""
        self.close_all()

    def _connect(self, **kwargs: Any) -> None:
        kwargs_to_paramiko = dict(
            look_for_keys=kwargs.pop('look_for_keys', True),
            username=kwargs.pop('username'),
            port=kwargs.pop('port', 22),
            allow_agent=False,
            timeout=kwargs.pop('timeout', None),
        )
        host = kwargs.pop('hostname', 'localhost')
        password = kwargs.pop('password')
        keepalive = kwargs.pop('keepalive', 5)
        if password:
            kwargs_to_paramiko['password'] = password
        self.raise_exceptions: bool = kwargs.pop('raise_exceptions', False)
        self.ssh_client = SSHClient()
        self.ssh_client.load_system_host_keys()
        self.ssh_client.connect(host, **kwargs_to_paramiko)
        self.client: OriginalSFTPClient = self.ssh_client.open_sftp()
        channel = self.client.get_channel()
        channel.settimeout(kwargs_to_paramiko['timeout'])
        channel.get_transport().set_keepalive(keepalive)
        # 'Extend' the SFTPClient class
        is_reconnect: bool = kwargs.pop('is_reconnect', False)
        members = inspect.getmembers(self.client, predicate=inspect.ismethod)
        self._log.debug('Dynamically adding methods from original SFTPClient')
        for method_name, method in members:
            if method_name[0:2] == '__' or method_name == '_log':
                self._log.debug('Ignoring %s()', method_name)
                continue
            if not is_reconnect and hasattr(self, method_name):
                raise AttributeError('Not overwriting property "{}". This '
                                     'version of Paramiko is not '
                                     'supported.'.format(method_name))
            self._log.debug('Adding method %s()', method_name)
            setattr(self, method_name, method)

    def close_all(self) -> None:
        """Close client and SSH client handles."""
        self.client.close()
        self.ssh_client.close()

    def clear_directory_cache(self) -> None:
        """Reset directory cache."""
        self._dircache = []

    def listdir_attr_recurse(self,
                             path: str = '.'
                             ) -> Iterator[Tuple[str, SFTPAttributes]]:
        """List directory attributes recursively."""
        for dir_attr in self.client.listdir_attr(path=path):
            is_dir = dir_attr.st_mode & 0o700 == 0o700
            if is_dir:
                try:
                    for x in self.listdir_attr_recurse(
                            path_join(path, dir_attr.filename)):
                        yield x
                except IOError as e:
                    if self.raise_exceptions:
                        raise e
            else:
                yield (
                    path_join(path, dir_attr.filename),
                    dir_attr,
                )

    def _get_callback(self, start_time: datetime,
                      _log: logging.Logger) -> Callable[[int, int], None]:
        def cb(tx_bytes: int, total_bytes: int) -> None:
            total_time = datetime.now() - start_time
            total_time_total_secs = total_time.total_seconds()
            total_time_s = floor(total_time_total_secs)

            if (total_time_s % LOG_INTERVAL) != 0:
                return

            nsize_tx = naturalsize(tx_bytes, binary=True, format='%.2f')
            nsize_total = naturalsize(total_bytes, binary=True, format='%.2f')

            speed_in_s = tx_bytes / total_time_total_secs
            speed_in_s = naturalsize(speed_in_s, binary=True, format='%.2f')

            _log.info('Downloaded %s / %s in %s (%s/s)', nsize_tx, nsize_total,
                      naturaldelta(datetime.now() - start_time), speed_in_s)

        return cb

    def mirror(self,
               path: str = '.',
               destroot: str = '.',
               keep_modes: bool = True,
               keep_times: bool = True,
               resume: bool = True) -> int:
        """
        Mirror a remote directory to a local location.

        path is the remote directory. destroot must be the location where
        destroot/path will be created (the path must not already exist).

        keep_modes and keep_times are boolean to ensure permissions and time
        are retained respectively.

        Pass resume=False to disable file resumption.
        """
        n = 0
        resume_seek = None
        cwd = self.getcwd()
        for _path, info in self.listdir_attr_recurse(path=path):
            if info.st_mode & 0o700 == 0o700:
                continue
            dest_path = path_join(destroot, dirname(_path))
            dest = path_join(dest_path, basename(_path))
            if dest_path not in self._dircache:
                try:
                    makedirs(dest_path)
                except OSError:
                    pass
                self._dircache.append(dest_path)
            if isdir(dest):
                continue
            try:
                with open(dest, 'rb'):
                    current_size = os.stat(dest).st_size
                    if current_size != info.st_size:
                        resume_seek = current_size
                        if resume:
                            self._log.info('Resuming file %s at %s bytes',
                                           dest, current_size)
                        raise IOError()  # ugly goto
            except IOError:
                while True:
                    try:
                        # Only size is used to determine complete-ness here
                        # Hash verification is in the util module
                        if resume_seek and resume:
                            read_tuples = []
                            n_reads = ceil((info.st_size - resume_seek) /
                                           self.MAX_PACKET_SIZE) - 1
                            n_left = ((info.st_size - resume_seek) %
                                      self.MAX_PACKET_SIZE)
                            offset = 0
                            for n in range(n_reads):
                                read_tuples.append((
                                    resume_seek + offset,
                                    self.MAX_PACKET_SIZE,
                                ))
                                offset += self.MAX_PACKET_SIZE
                            read_tuples.append((
                                resume_seek + offset,
                                n_left,
                            ))
                            with self.client.open(_path) as sftp_file:
                                with open(dest, 'ab') as f:
                                    f.seek(resume_seek)
                                    resume_seek = None
                                    for chunk in sftp_file.readv(read_tuples):
                                        f.write(chunk)
                        else:
                            dest = realpath(dest)
                            self._log.info('Downloading %s -> %s', _path, dest)
                            start_time = datetime.now()
                            self.client.get(_path, dest)
                            self._get_callback(start_time,
                                               self._log)(info.st_size,
                                                          info.st_size)

                        # Do not count files that were already downloaded
                        n += 1

                        break
                    except (socket.timeout, SFTPError) as e:
                        # Resume at position - 10 bytes
                        resume_seek = os.stat(dest).st_size - 10
                        if isinstance(e, socket.timeout):
                            self._log.error('Connection timed out')
                        else:
                            self._log.error('%s', e)
                        if resume:
                            self._log.info('Resuming GET %s at %s bytes',
                                           _path, resume_seek)
                        else:
                            self._log.debug(
                                'Not resuming (resume = %s, exception: %s)',
                                resume, e)
                            raise e
                        self._log.debug('Re-establishing connection')
                        self.original_arguments['is_reconnect'] = True
                        self._connect(**self.original_arguments)
                        if cwd:
                            cast(OriginalSFTPClient, self).chdir(cwd)
            # Okay to fix existing files even if they are already downloaded
            try:
                if keep_modes:
                    chmod(dest, info.st_mode)
                if keep_times:
                    utime(dest, (
                        info.st_atime,
                        info.st_mtime,
                    ))
            except IOError:
                pass
        return n

    def __str__(self) -> str:
        """Return string representation."""
        return f'{self.client} (wrapped by {__name__}.SFTPClient)'



from hashlib import sha1
from io import BytesIO as StringIO
from os import close as close_fd, remove as rm, rmdir, write as write_fd
from os.path import basename, dirname
from random import SystemRandom
from tempfile import mkdtemp, mkstemp
from typing import List, Optional
import sys
import unittest

from benc import encode as bencode

from xirvik.util import VerificationError, verify_torrent_contents

random = SystemRandom()


def create_random_data(size: int):
    return bytearray(random.getrandbits(8) for _ in range(size))


class TempFilesMixin:
    _temp_files: List[str] = []

    def tearDown(self):
        for x in self._temp_files:
            try:
                rm(x)
            except OSError as e:
                if e.errno == 2:
                    continue
                else:
                    print(str(e), file=sys.stderr)

    def _mktemp(self,
                contents: Optional[bytes] = None,
                prefix: str = 'test-',
                dir: Optional[str] = None):
        fd, name = mkstemp(prefix=prefix, dir=dir)
        write_fd(fd, contents or b'')
        close_fd(fd)

        self._temp_files.append(name)

        return name


class TestTorrentVerification(TempFilesMixin, unittest.TestCase):
    FILE_SIZE = 2509
    PIECE_LENGTH = 256

    def setUp(self):
        """ A torrent generator! """
        self.torrent_data_path = mkdtemp(prefix='test-torrent-verification-')
        self.torrent_name = basename(self.torrent_data_path)

        all_data = create_random_data(self.FILE_SIZE * 2)
        pieces = b''

        self.file1 = self._mktemp(contents=all_data[0:self.FILE_SIZE],
                                  dir=self.torrent_data_path)
        self.file2 = self._mktemp(contents=all_data[self.FILE_SIZE:],
                                  dir=self.torrent_data_path)

        for i in range(0, self.FILE_SIZE * 2, self.PIECE_LENGTH):
            s = sha1()
            s.update(all_data[i:i + self.PIECE_LENGTH])
            pieces += s.digest()

        self.torrent_data_dict = {
            b'announce': b'https://fake.com',
            b'info': {
                b'name':
                self.torrent_name.encode('utf-8'),
                b'piece length':
                self.PIECE_LENGTH,
                b'pieces':
                pieces,
                b'files': [
                    {
                        b'length': self.FILE_SIZE,
                        b'path': [basename(self.file1).encode('utf-8')],
                    },
                    {
                        b'length': self.FILE_SIZE,
                        b'path': [basename(self.file2).encode('utf-8')],
                    },
                ],
            }
        }
        self.torrent_data = bencode(self.torrent_data_dict)

        self.torrent_file_path = self._mktemp(contents=self.torrent_data)

    def tearDown(self):
        super(TestTorrentVerification, self).tearDown()
        rmdir(self.torrent_data_path)

    def test_verify_torrent_contents_string(self):
        verify_torrent_contents(self.torrent_data,
                                dirname(self.torrent_data_path))

    def test_verify_torrent_contents_filepath(self):
        verify_torrent_contents(self.torrent_file_path,
                                dirname(self.torrent_data_path))

    def test_verify_torrent_contents_stringio(self):
        verify_torrent_contents(StringIO(self.torrent_data),
                                dirname(self.torrent_data_path))

    def test_verify_torrent_contents_invalid_path(self):
        with self.assertRaises(IOError):
            verify_torrent_contents(self.torrent_data,
                                    dirname(self.torrent_data_path) + 'junk')

    def test_verify_torrent_contents_file_missing(self):
        rm(self.file2)
        with self.assertRaises(VerificationError):
            verify_torrent_contents(self.torrent_data,
                                    dirname(self.torrent_data_path))

    def test_verify_torrent_contents_keyerror(self):
        del self.torrent_data_dict[b'info'][b'files'][0][b'path']
        self.torrent_data = bencode(self.torrent_data_dict)

        with self.assertRaises(KeyError):
            verify_torrent_contents(self.torrent_data,
                                    dirname(self.torrent_data_path))

    def test_verify_torrent_contents_keyerror2(self):
        del self.torrent_data_dict[b'info']
        self.torrent_data = bencode(self.torrent_data_dict)

        with self.assertRaises(KeyError):
            verify_torrent_contents(self.torrent_data,
                                    dirname(self.torrent_data_path))

    def test_verify_torrent_contents_bad_compare(self):
        with open(self.file2, 'w') as f:
            f.write('junk\n')

        with self.assertRaises(VerificationError):
            verify_torrent_contents(self.torrent_file_path,
                                    dirname(self.torrent_data_path))


class TestSingleFileTorrentVerification(TempFilesMixin, unittest.TestCase):
    FILE_SIZE = 2509
    PIECE_LENGTH = 256

    def setUp(self):
        all_data = create_random_data(self.FILE_SIZE)
        self.file1 = self._mktemp(contents=all_data)
        self.torrent_data_path = dirname(self.file1)

        pieces = b''
        for i in range(0, self.FILE_SIZE, self.PIECE_LENGTH):
            s = sha1()
            s.update(all_data[i:i + self.PIECE_LENGTH])
            pieces += s.digest()

        self.torrent_data_dict = {
            b'announce': b'https://fake.com',
            b'info': {
                b'name': self.file1.encode('utf-8'),
                b'piece length': self.PIECE_LENGTH,
                b'pieces': pieces,
            }
        }
        self.torrent_data = bencode(self.torrent_data_dict)

        self.torrent_file_path = self._mktemp(contents=self.torrent_data)

    def test_verify_torrent_contents_string(self):
        verify_torrent_contents(self.torrent_data, self.torrent_data_path)

    def test_verify_torrent_contents_filepath(self):
        verify_torrent_contents(self.torrent_file_path, self.torrent_data_path)

    def test_verify_torrent_contents_stringio(self):
        verify_torrent_contents(StringIO(self.torrent_data),
                                self.torrent_data_path)

    def test_verify_torrent_contents_file_missing(self):
        rm(self.file1)
        with self.assertRaises(IOError):
            verify_torrent_contents(self.torrent_data, self.torrent_data_path)

    def test_verify_torrent_contents_bad_compare(self):
        with open(self.file1, 'w') as f:
            f.write('junk\n')

        with self.assertRaises(VerificationError):
            verify_torrent_contents(self.torrent_file_path,
                                    self.torrent_data_path)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""Organise torrents based on labels assigned in ruTorrent."""
from time import sleep
from typing import Any, Callable, Dict, Mapping, Tuple, Union
import sys

from requests.exceptions import HTTPError
import argcomplete

from ..client import ruTorrentClient
from .util import common_parser, setup_logging_stdout

PREFIX = '/torrents/{}/_completed'


def _base_path_check(
        username: str) -> Callable[[Tuple[Any, Mapping[str, str]]], bool]:
    def bpc(hash_info: Tuple[Any, Mapping[str, str]]) -> bool:
        _, info = hash_info
        move_to = '{}/{}'.format(PREFIX.format(username),
                                 str.lower(info['custom1']))
        return not info['base_path'].startswith(move_to)

    return bpc


def _key_check(hash_info: Tuple[Any, Mapping[str, Union[bool, int]]]) -> bool:
    _, info = hash_info
    return not info['is_hash_checking'] and info['left_bytes'] == 0


def main() -> int:
    """Entry point."""
    parser = common_parser()
    parser.add_argument(
        '-c',
        '--completed-dir',
        default='_completed',
        help='Top directory where moved torrent data will be placed')
    parser.add_argument(
        '-t',
        '--sleep-time',
        default=10,
        type=int,
        help=('Time to sleep in seconds at certain times during this batch '
              'of requests'))
    parser.add_argument(
        '-l',
        '--lower-label',
        action='store_true',
        help='Call lower() on labels used to make directory names')
    parser.add_argument('--ignore-labels',
                        nargs='+',
                        default=[],
                        help='List of label names to ignore')
    argcomplete.autocomplete(parser)
    args = parser.parse_args()
    log = setup_logging_stdout(verbose=args.verbose)
    assert log is not None
    client = ruTorrentClient(args.host[0],
                             name=args.username,
                             password=args.password,
                             max_retries=args.max_retries,
                             netrc_path=args.netrc)
    username = client.name
    try:
        torrents = list(client.list_torrents_dict().items())
    except (ValueError, HTTPError):
        log.error('Connection failed on list_torrents() call')
        return 1
    count = 0
    assert username is not None
    hash_: str
    info: Dict[str, str]
    for hash_, info in list(
            filter(_base_path_check(username), filter(_key_check, torrents))):
        label = info['custom1']
        if label in args.ignore_labels:
            continue
        if args.lower_label:
            label = label.lower()
        move_to = '{}/{}'.format(PREFIX.format(username), label)
        log.info('Moving %s to %s/', info['name'], move_to)
        client.move_torrent(hash_, move_to)
        count += 1
        if count > 0 and (count % 10) == 0:
            sleep(args.sleep_time)
    return 0


if __name__ == '__main__':
    sys.exit(main())



"""Client for ruTorrent."""
from cgi import parse_header
from datetime import datetime
from netrc import netrc
from os.path import expanduser
from typing import (Any, Callable, Dict, Iterable, Iterator, Mapping, Optional,
                    Sequence, Tuple, Union, cast)
from urllib.parse import quote
import logging
import re
import ssl
import xmlrpc.client as xmlrpc

from cached_property import cached_property
from requests.adapters import HTTPAdapter
from requests_futures.sessions import FuturesSession
from urllib3.util import Retry
import requests

from .typing import TorrentDict

__all__ = (
    'LOG_NAME',
    'TORRENT_PATH_INDEX',
    'UnexpectedruTorrentError',
    'ruTorrentClient',
)

#: Name used in logger.
LOG_NAME = 'xirvik.rutorrent'
TORRENT_FILE_DOWNLOAD_STRATEGY_LEADING_CHUNK_FIRST = 1
TORRENT_FILE_DOWNLOAD_STRATEGY_NORMAL = 0
TORRENT_FILE_DOWNLOAD_STRATEGY_TRAILING_CHUNK_FIRST = 2
TORRENT_FILE_PRIORITY_DONT_DOWNLOAD = 0
TORRENT_FILE_PRIORITY_HIGH = 2
TORRENT_FILE_PRIORITY_NORMAL = 1
TORRENT_LABEL_INDEX = 14
#: Index of the torrent information list that has the path
TORRENT_PATH_INDEX = 25
TORRENT_PIECE_SIZE_INDEX = 13


class UnexpectedruTorrentError(Exception):
    """Raised when an unexpected error occurs."""


class ruTorrentClient:
    """
    ruTorrent client class.

    Reference on RPC returned fields: https://goo.gl/DvmW4c
    """
    def __init__(self,
                 host: str,
                 name: Optional[str] = None,
                 password: Optional[str] = None,
                 max_retries: int = 10,
                 netrc_path: Optional[str] = None):
        """
        Construct a ruTorrent client.

        Host should be the hostname with no protocol.

        If no name and no password are passed, ~/.netrc will be searched with
        the host provided. The path can be overridden with the netrc_path
        argument.

        max_retries is used as an argument for urllib3's Retry() class.
        """
        if not name and not password:
            if not netrc_path:
                netrc_path = expanduser('~/.netrc')
            name, _, password = cast(Tuple[str, ...],
                                     netrc(netrc_path).authenticators(host))
        self.name = name
        self.password = password
        self.host = host
        retry = Retry(connect=max_retries,
                      read=max_retries,
                      redirect=False,
                      backoff_factor=1)
        self._log = logging.getLogger(LOG_NAME)
        self._http_adapter = HTTPAdapter(max_retries=retry)
        self._session = requests.Session()
        self._session.mount('http://', self._http_adapter)
        self._session.mount('https://', self._http_adapter)

        uri = 'https://{}:{}@{}/rtorrent/plugins/multirpc/action.php'.format(
            self.name, self.password, self.host)
        self._xmlrpc_proxy = xmlrpc.ServerProxy(
            uri, context=ssl._create_unverified_context())

    @cached_property
    def http_prefix(self) -> str:
        """Return HTTP URI for the host."""
        return f'https://{self.host:s}'

    @cached_property
    def multirpc_action_uri(self) -> str:
        """Return HTTP multirpc/action.php URI for the host."""
        return f'{self.http_prefix}/rtorrent/plugins/multirpc/action.php'

    @cached_property
    def datadir_action_uri(self) -> str:
        """Return HTTP datadir/action.php URI for the host."""
        return f'{self.http_prefix}/rtorrent/plugins/datadir/action.php'

    @cached_property
    def _add_torrent_uri(self) -> str:
        return f'{self.http_prefix}/rtorrent/php/addtorrent.php?'

    @cached_property
    def auth(self) -> Tuple[Optional[str], Optional[str]]:
        """Return basic authentication credentials."""
        return (
            self.name,
            self.password,
        )

    def add_torrent(self, filepath: str, start_now: bool = True) -> None:
        """Add a torrent. Use start_now=False to start paused."""
        with open(filepath, 'rb') as f:
            self._session.post(
                self._add_torrent_uri,
                data=dict(
                    torrents_start_stopped='on') if not start_now else {},
                auth=self.auth,
                files=dict(torrent_file=f)).raise_for_status()

    def list_torrents(self) -> Mapping[str, Sequence[Any]]:
        """
        List torrents as they come from ruTorrent.

        Return a dictionary with torrent hashes as the key. The values
        are lists similar to the columns in ruTorrent's main view.

        For a more detailed dictionary, use list_torrents_dict().
        """
        r = self._session.post(self.multirpc_action_uri,
                               data={
                                   'mode': 'list',
                                   'cmd': 'd.custom=addtime',
                               },
                               auth=self.auth)
        r.raise_for_status()
        ret = r.json()['t']
        if isinstance(ret, list):
            raise ValueError('Unexpected type from API')
        return cast(Mapping[str, Sequence[Any]], ret)

    def list_torrents_dict(self) -> Mapping[str, TorrentDict]:
        """
        Get all torrent information.

        Return a dictionary of dictionaries with the hash of the torrent as
        the key. The fields will be:

        - is_open - boolean, ???
        - is_hash_checking - boolean, if the torrent is hash checking
        - is_hash_checked - boolean, if the torrent is already hash checked
        - state - integer, state of the torrent (out of some enumeration)
        - name - string, Name of the torrent
        - size_bytes - integer, size of the torrent in bytes
        - completed_chunks - integer, completed number of chunks
        - size_chunks - integer, number of chunks in the torrent
        - bytes_done - integer, bytes completed downloading
        - up_total - integer, amount of bytes uploaded
        - ratio - float, Ratio
        - up_rate - integer, upload rate in bytes per second
        - down_rate - integer, download rate in bytes per second
        - chunk_size - integer, size of the chunks
        - custom1 - string, usually label, blank string if nothing is set
        - peers_accounted - integer, total number of peers
        - peers_not_connected - integer, number of inactive peers
        - peers_connected - integer, numbers of active peers
        - peers_complete - integer, number of peers who have the completed
                                    torrent
        - left_bytes - integer, number of bytes left to download
        - priority - integer, out of some enumeration
        - state_changed - datetime, last time the torrent was modified
        - skip_total - integer, ???
        - hashing - boolean, if the torrent is hashing
        - chunks_hashed - integer, number of chunks hashed
        - base_path - string, path to the torrent files
        - creation_date - None or datetime
        - tracker_focus - integer, ???
        - is_active - if the torrent is active
        - message - string, ???
        - custom2 - string, ???
        - free_diskspace - integer, disk space available on the server
        - is_private - boolean, if the torrent is private
        - is_multi_file - boolean, if the torrent contains multiple files
        """
        fields = (
            'is_open',
            'is_hash_checking',
            'is_hash_checked',
            'state',
            'name',
            'size_bytes',
            'completed_chunks',
            'size_chunks',
            'bytes_done',
            'up_total',
            'ratio',
            'up_rate',
            'down_rate',
            'chunk_size',
            'custom1',
            'peers_accounted',
            'peers_not_connected',
            'peers_connected',
            'peers_complete',
            'left_bytes',
            'priority',
            'state_changed',
            'skip_total',
            'hashing',
            'chunks_hashed',
            'base_path',
            'creation_date',
            'tracker_focus',
            'is_active',
            'message',
            'custom2',
            'free_diskspace',
            'is_private',
            'is_multi_file',
        )
        ret: Dict[str, Dict[str, Any]] = {}
        for hash_, torrent in self.list_torrents().items():
            ret[hash_] = {}
            value: Any
            for i, value in enumerate(torrent):
                try:
                    if fields[i].startswith('is_') or fields[i] == 'hashing':
                        value = True if value == '1' else False
                    elif fields[i] == 'state_changed':
                        value = datetime.fromtimestamp(float(value))
                    elif fields[i] == 'creation_date':
                        try:
                            fvalue: Optional[float] = float(value)
                        except ValueError:
                            fvalue = None
                        if fvalue:
                            value = datetime.fromtimestamp(float(fvalue))
                        else:
                            value = None
                    elif fields[i] == 'ratio':
                        value = int(value) / 1000.0
                    else:
                        re1 = r'^[0-9]+(\.[0-9]+)?'
                        re2 = r'^\.[0-9]+'
                        if re.match(re1, value) or re.match(re2, value):
                            cons = int if "." not in value else float
                            try:
                                value = cons(value)
                            except ValueError:
                                pass
                    ret[hash_][fields[i]] = value
                except IndexError:
                    continue
        return cast(Mapping[str, TorrentDict], ret)

    def get_torrent(self, hash_: str) -> Tuple[requests.Response, str]:
        """
        Prepare to get a torrent file given a hash.

        Return tuple Request object and the file name string.
        """
        source_torrent_uri = (f'{self.http_prefix}/rtorrent/plugins/source/'
                              f'action.php?hash={hash_}')
        r = self._session.get(source_torrent_uri, auth=self.auth, stream=True)
        r.raise_for_status()
        fn = parse_header(r.headers['content-disposition'])[1]['filename']
        return r, fn

    def get_torrents_futures(
        self,
        hashes: Iterable[str],
        session: Optional[FuturesSession] = None,
        background_callback: Callable[[Any], Any] = None
    ) -> Iterator[FuturesSession]:
        """
        Similar to get_torrent() but uses requests_futures.

        Pass a list of hashes, optionally a session and a callback.

        Yields the GET future request for each hash.
        """
        if not session:
            session = FuturesSession(max_workers=4)
        for hash_ in hashes:
            yield session.get((f'{self.http_prefix}/rtorrent/plugins/source/'
                               f'action.php?hash={hash_}'),
                              background_callback=background_callback)

    def move_torrent(self,
                     hash_: str,
                     target_dir: str,
                     fast_resume: bool = True) -> None:
        """
        Move a torrent's files to somewhere else on the server.

        target_dir must be a valid and usable directory.
        """
        r = self._session.post(self.datadir_action_uri,
                               data={
                                   'hash': hash_,
                                   'datadir': target_dir,
                                   'move_addpath': '1',
                                   'move_datafiles': '1',
                                   'move_fastresume':
                                   '1' if fast_resume else '0',
                               },
                               auth=self.auth)
        r.raise_for_status()
        json = r.json()
        if 'errors' in json and json['errors']:
            raise UnexpectedruTorrentError(str(json['errors']))

    def set_label_to_hashes(self, **kwargs: Any) -> None:
        """
        Set a label to a list of info hashes. The label can be a new label.

        To remove a label, pass an empty string as the `label` keyword
        argument.

        Example use:
            client.set_labels(hashes=[hash_1, hash_2], label='my new label')
        """
        # The way to set a label to multiple torrents is to specify the hashes
        # using hash=, then the v parameter as many times as there are hashes,
        # and then the s=label for as many times as there are hashes.
        #
        # Example:
        #    mode=setlabel&hash=...&hash=...&v=label&v=label&s=label&s=label
        #
        # This method builds this string out since Requests can take in a byte
        # string as POST data (and you cannot set a key twice in a dictionary).
        hashes = kwargs.pop('hashes', [])
        label = kwargs.pop('label', None)
        allow_recursive_fix = kwargs.pop('allow_recursive_fix', True)
        recursion_limit = kwargs.pop('recursion_limit', 5)
        recursion_attempt = kwargs.pop('recursion_attempt', 0)
        if not hashes or not label:
            raise TypeError('"hashes" (list) and "label" (str) keyword '
                            'arguments are required')
        data = b'mode=setlabel'
        for hash_ in hashes:
            data += '&hash={}'.format(hash_).encode()
        data += '&v={}'.format(label).encode() * len(hashes)
        data += b'&s=label' * len(hashes)
        self._log.debug('set_labels() with data: %s', data.decode('utf-8'))
        r = self._session.post(self.multirpc_action_uri,
                               data=data,
                               auth=self.auth)
        r.raise_for_status()
        json = r.json()
        # This may not be an error, but sometimes just `[]` is returned
        # Even with the retries, sometimes all but one torrent gets a label
        if len(json) != len(hashes):
            self._log.warning(
                'JSON returned should have been an array with '
                'same length as hashes list passed in: %s', json)
            if allow_recursive_fix and recursion_attempt < recursion_limit:
                recursion_attempt += 1
                self._log.info('Attempting label again '
                               '(%d out of %d)', recursion_attempt,
                               recursion_limit)

                data = b'mode=setlabel'
                new_hashes = []
                for hash_, v in self.list_torrents().items():
                    if hash_ not in hashes or v[TORRENT_LABEL_INDEX].strip():
                        continue
                    data += '&hash={}'.format(hash_).encode()
                    new_hashes.append(hash_)
                if not new_hashes:
                    self._log.debug('Found no torrents to correct')
                    return
                self.set_label_to_hashes(hashes=new_hashes,
                                         label=label,
                                         recursion_limit=recursion_limit,
                                         recursion_attempt=recursion_attempt)
            else:
                self._log.warning('Passed recursion limit for label fix')

    def set_label(self, label: str, hash_: str) -> None:
        """Set a label to a torrent hash."""
        self.set_label_to_hashes(hashes=[hash_], label=label)

    def list_files(self, hash_: str) -> Iterator[Sequence[Union[int, str]]]:
        """
        List files for a given torrent hash.

        Returns a generator of lists with fields in this order:
        - name
        - total number of pieces
        - number of pieces downloaded
        - size in bytes
        - priority
        - download strategy
        - ??? (some integer)

        for info in list_files():
            for name, pieces, pieces_dl, size, dlstrat, _ in info:
        """
        cmds: Union[str, Tuple[str, ...]] = (
            quote('f.prioritize_first='),
            quote('f.prioritize_last='),
        )
        query = 'mode=fls&hash={}'.format(hash_).encode()
        cmds = '&'.join(['cmd={}'.format(x) for x in cmds])
        query += b'&'
        query += cmds.encode()
        r = self._session.post(self.multirpc_action_uri,
                               data=query,
                               auth=self.auth)
        r.raise_for_status()
        for x in r.json():
            # Fix the numeric values which come as strings
            x[1] = int(x[1])  # total number of pieces
            x[2] = int(x[2])  # downloaded pieces
            x[3] = int(x[3])  # size in bytes
            x[4] = int(x[4])  # priority ID
            x[5] = int(x[5])  # download strategy ID
            x[6] = int(x[6])  # ??
            yield x

    def delete(self, hash_: str) -> None:
        """
        Delete a torrent and its files by hash. Use the remove() method to
        remove the torrent but keep the data.

        Returns if successful. Faults are converted to xmlrpc.Fault exceptions.
        """
        mc = xmlrpc.MultiCall(self._xmlrpc_proxy)
        getattr(mc, 'd.custom5.set')(hash_, '1')
        getattr(mc, 'd.delete_tied')(hash_)
        getattr(mc, 'd.erase')(hash_)
        for x in mc().results:
            try:
                raise xmlrpc.Fault(x['faultCode'], x['faultString'])
            except (TypeError, KeyError):
                pass

    def remove(self, hash_: str) -> None:
        """
        Remove a torrent from the client but keep the data. Use the delete()
        method to remove and delete the torrent data.

        Returns if successful. Can raise a Requests exception.
        """
        self._session.post(self.multirpc_action_uri,
                           data=dict(mode='remove', hash=hash_),
                           auth=self.auth).raise_for_status()

    def stop(self, hash_: str) -> None:
        """
        Stop a torrent by hash.

        Returns if successful. Can raise a Requests exception.
        """
        self._session.post(self.multirpc_action_uri,
                           data=dict(mode='stop', hash=hash_),
                           auth=self.auth).raise_for_status()

    def add_torrent_url(self, url: str) -> None:
        """Add a torrent via a publicly accessible URI."""
        self._session.post(self._add_torrent_uri,
                           data=dict(url=url),
                           auth=self.auth).raise_for_status()



from .simple import (add_ftp_user, authorize_ip, delete_ftp_user, fix_rtorrent,
                     mirror_main, start_torrents)

__all__ = [
    'add_ftp_user', 'authorize_ip', 'delete_ftp_user', 'fix_rtorrent',
    'mirror_main', 'start_torrents'
]



"""Typing helpers."""
from datetime import datetime
from typing import Callable, Optional, TypeVar

from typing_extensions import TypedDict

__all__ = ('Method0', 'Method1', 'TorrentDict')

T = TypeVar('T')
U = TypeVar('U')
V = TypeVar('V')
Method0 = Callable[[T], V]
Method1 = Callable[[T, U], V]


class TorrentDict(TypedDict):
    base_path: str
    bytes_done: int
    chunk_size: int
    chunks_hashed: int
    completed_chunks: int
    creation_date: Optional[datetime]
    custom1: str
    custom2: str
    down_rate: int
    free_diskspace: int
    hashing: bool
    is_active: int
    is_hash_checked: bool
    is_hash_checking: bool
    is_multi_file: bool
    is_open: bool
    is_private: bool
    left_bytes: int
    message: str
    peers_accounted: int
    peers_connected: int
    peers_not_connected: int
    priority: int
    ratio: float
    size_bytes: int
    size_chunks: int
    skip_total: int
    state: int
    state_changed: datetime
    tracker_focus: int
    up_rate: int
    up_total: int



from os import close as close_fd, remove as rm, write as write_fd
from tempfile import mkstemp
from typing import List, Optional
import unittest

from requests.exceptions import HTTPError
import requests_mock

from xirvik.client import (TORRENT_FILE_DOWNLOAD_STRATEGY_NORMAL,
                           TORRENT_FILE_PRIORITY_NORMAL,
                           UnexpectedruTorrentError, ruTorrentClient)


def isfile(filepath: str) -> bool:
    try:
        with open(filepath, 'rb'):
            return True
    except IOError:
        pass

    return False


class TestRuTorrentClient(unittest.TestCase):
    _temp_files: List[str] = []

    def tearDown(self):
        for x in self._temp_files:
            try:
                rm(x)
            except OSError:
                pass

    def _mktemp(self,
                contents: Optional[bytes] = None,
                prefix: str = 'test-rutorrent-client-'):
        fd, name = mkstemp(prefix=prefix)
        write_fd(fd, contents or b'')
        close_fd(fd)

        self._temp_files.append(name)

        return name

    def test_netrc(self):
        netrc_line = ('machine hostname-test.com login a password '
                      'bbbb\n').encode('utf-8')
        name = self._mktemp(contents=netrc_line)

        client = ruTorrentClient('hostname-test.com', netrc_path=name)
        self.assertEqual('a', client.name)
        self.assertEqual('bbbb', client.password)
        self.assertEqual('hostname-test.com', client.host)

        self.assertEqual('a', client.auth[0])
        self.assertEqual('bbbb', client.auth[1])

    def test_http_prefix(self):
        client = ruTorrentClient('hostname-test.com', 'a', 'b')
        self.assertEqual('https://hostname-test.com', client.http_prefix)

    @requests_mock.Mocker()
    def test_add_torrent_bad_status(self, m):
        client = ruTorrentClient('hostname-test.com', 'a', 'b')
        torrent = self._mktemp('torrent file fake'.encode('utf-8'))

        m.post(client._add_torrent_uri, status_code=400)
        with self.assertRaises(HTTPError):
            client.add_torrent(torrent)

    @requests_mock.Mocker()
    def test_list_torrents_bad_status(self, m: requests_mock.Mocker):
        client = ruTorrentClient('hostname-test.com', 'a', 'b')
        m.post(client.multirpc_action_uri, status_code=400)

        with self.assertRaises(HTTPError):
            client.list_torrents()

    @requests_mock.Mocker()
    def test_list_torrents(self, m: requests_mock.Mocker):
        client = ruTorrentClient('hostname-test.com', 'a', 'b')
        m.post(client.multirpc_action_uri,
               json=dict(
                   t={
                       'hash here': ['1', '0', '1', '1', 'name of torrent?'],
                   },
                   cid=92385,
               ))

        self.assertEqual(client.list_torrents()['hash here'][4],
                         'name of torrent?')

    @requests_mock.Mocker()
    def test_get_torrent(self, m: requests_mock.Mocker):
        client = ruTorrentClient('hostname-test.com', 'a', 'b')
        uri = ('{}/rtorrent/plugins/source/action.php'
               '?hash=hash_of_torrent').format(client.http_prefix)
        m.get(uri,
              headers={
                  'content-disposition': 'attachment; '
                  'filename=test.torrent'
              })
        _, fn = client.get_torrent('hash_of_torrent')

        self.assertEqual('test.torrent', fn)

    @requests_mock.Mocker()
    def test_list_files(self, m: requests_mock.Mocker):
        client = ruTorrentClient('hostname-test.com', 'a', 'b')

        m.post(client.multirpc_action_uri,
               json=[
                   ['name of file', '14', '13', '8192', '1', '0', '0'],
               ])

        files = list(client.list_files('testhash'))
        self.assertEqual('name of file', files[0][0])
        self.assertEqual(14, files[0][1])
        self.assertEqual(13, files[0][2])
        self.assertNotEqual('8192', files[0][3])
        self.assertEqual(8192, files[0][3])
        self.assertEqual(TORRENT_FILE_PRIORITY_NORMAL, files[0][4])
        self.assertEqual(TORRENT_FILE_DOWNLOAD_STRATEGY_NORMAL, files[0][5])

    @requests_mock.Mocker()
    def test_set_label_to_hashes_recursion_limit_5(self,
                                                   m: requests_mock.Mocker):
        client = ruTorrentClient('hostname-test.com', 'a', 'b')
        hashes = ['hash1', 'hash2']
        label = 'my new label'
        list_torrents_json = dict(
            t={
                'hash1': [
                    '1', '0', '1', '1', 'torrent name', '250952849', '958',
                    '958', '250952849', '357999402', '1426', '0', '0',
                    '262144', ''
                ],
                'hash2': [
                    '1', '0', '1', '1', 'torrent name2', '250952849', '958',
                    '958', '250952849', '357999402', '1426', '0', '0',
                    '262144', ''
                ],
            },
            cid=92983,
        )
        responses = [
            dict(json=[]),
            dict(json=list_torrents_json),
            dict(json=[]),
            dict(json=list_torrents_json),
            dict(json=[]),
            dict(json=list_torrents_json),
            dict(json=[]),
            dict(json=list_torrents_json),
            dict(json=[]),
            dict(json=list_torrents_json),
            dict(json=[]),
        ]

        m.post(client.multirpc_action_uri, responses)
        client.set_label_to_hashes(hashes=hashes,
                                   label=label,
                                   recursion_limit=5)

        with self.assertRaises(TypeError):
            client.set_label_to_hashes()

    @requests_mock.Mocker()
    def test_set_label(self, m: requests_mock.Mocker):
        client = ruTorrentClient('hostname-test.com', 'a', 'b')
        list_torrents_json = dict(
            t=dict(hash1=[
                '1', '0', '1', '1', 'torrent name', '250952849', '958', '958',
                '250952849', '357999402', '1426', '0', '0', '262144', 'a label'
            ], ),
            cid=92983,
        )
        responses = [
            dict(json=[]),
            dict(json=list_torrents_json),
        ]

        m.post(client.multirpc_action_uri, responses)
        client.set_label('hash1', 'a label')

    @requests_mock.Mocker()
    def test_move_torrent(self, m: requests_mock.Mocker):
        client = ruTorrentClient('hostname-test.com', 'a', 'b')

        m.post(client.datadir_action_uri, json=[], status_code=400)
        with self.assertRaises(HTTPError):
            client.move_torrent('hash1', 'newplace')

        m.post(client.datadir_action_uri, json={'errors': ['some error']})
        with self.assertRaises(UnexpectedruTorrentError):
            client.move_torrent('hash1', 'newplace')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
from functools import partial
from time import sleep
from typing import (Any, Callable, Hashable, Iterator, List, Mapping, Optional,
                    Sequence, Tuple)
import logging
import sys

from ..client import ruTorrentClient
from .util import common_parser, setup_logging_stdout

PREFIX = '/torrents/{}/_completed-not-active'
BAD_MESSAGES = (
    'unregistered torrent',
    "couldn't connect to server",
    'server returned nothing',
)
log: Optional[logging.Logger] = None


def get_name(x: Mapping[str, str]) -> str:
    return x['name']


def get_message(x: Mapping[str, str]) -> str:
    return x['message']


def get_label(x: Mapping[str, str]) -> str:
    return x['custom1']


def get_label_lower(x: Mapping[str, str]) -> str:
    return get_label(x).lower()


def is_hash_checking(x: Mapping[str, bool]) -> bool:
    return x['is_hash_checking']


def complement(x: Callable[..., Any]) -> Callable[..., bool]:
    def z(*args: Any, **kwargs: Any) -> bool:
        return not x(*args, **kwargs)

    return z


def has_one(look_for_list: Sequence[Hashable],
            val: Sequence[Hashable]) -> bool:
    for s in look_for_list:
        if s in val:
            return True
    return False


def has_bad_message(x: Mapping[str, str]) -> bool:
    return has_one(BAD_MESSAGES, x['message'])


def left_bytes_gt_0(x: Mapping[str, int]) -> bool:
    return x['left_bytes'] > 0


# def make_move_to(prefix, label):
#     return '{}/{}'.format(prefix, label)

# make_move_to_prefix = partial(make_move_to, PREFIX)
# mmtp = compose(make_move_to_prefix, get_label_lower)

# def rem0(div, x):
#     return compose(partial(equals, 0), partial_right(mod, div))(x)

# rem_zero_ten = partial(rem0, 10)

# def _should_skip(move_to):
#     return any_pass(is_hash_checking, left_bytes_gt_0,
#                     base_path_check(move_to))

# should_skip = compose(_should_skip, mmtp)

# rem_zero_ten_not_zero = cond((
#     (lambda i: i <= 0, f),
#     (rem_zero_ten, t),
#     (t, t),
# ))


def any_pass(*funcs: Callable[..., bool]) -> Callable[[Any], bool]:
    def ret(x: Any) -> bool:
        for pred in funcs:
            if pred(x):
                return True
        return False

    return ret


# def starts_with(search: str, x: str) -> bool:
#     return str.startswith(x, search)

# def base_path_check(move_to: str) -> bool:
#     return compose(partial(starts_with, move_to), partial(get, 'base_path'))

# def do_if(pred: Callable[..., Any], args: Sequence[Any],
#           func: Callable[..., Any], func_args: Sequence[Any]):
#     if pred(*args):
#         return func(*func_args)


def main() -> int:
    global log
    parser = common_parser()
    parser.add_argument('-a', '--ignore-ratio', action='store_true')
    parser.add_argument('-t', '--sleep-time', default=10, type=int)
    args = parser.parse_args()
    log = setup_logging_stdout(verbose=args.verbose)
    assert log is not None
    client = ruTorrentClient(args.host[0],
                             name=args.username,
                             password=args.password,
                             max_retries=args.max_retries,
                             netrc_path=args.netrc)
    prefix = PREFIX.format(args.username)
    to_delete: List[Tuple[str, str]] = []

    items: Iterator[Tuple[str, Mapping[str, Any]]] = filter(
        lambda x: has_bad_message(x[1]),
        client.list_torrents().items())

    count = 0
    for hash_, info in items:
        log.info('Stopping %s', info['name'])
        client.stop(hash_)
        count += 1
        if count > 0 and (count % 10) == 0:
            sleep(args.sleep_time)

    should_process_ = compose(complement(bool), partial(get, 'is_active'))
    items = filter(
        lambda: any_pass(is_hash_checking, left_bytes_gt_0,
                         base_path_check(move_to)), filter())

    count = 0
    for hash_, info in items:
        move_to = make_move_to(prefix, info['custom1'].lower())
        name = get_name(info)
        to_delete.append((hash_, name))
        log.info('Moving %s to %s/', name, move_to)
        client.move_torrent(hash_, move_to)
        client.stop(hash_)
        count += 1
        if count > 0 and (count % 10) == 0:
            sleep(args.sleep_time)

    count = 0
    for hash_, name in to_delete:
        log.info('Removing torrent "%s" (without deleting data)', name)
        client.remove(hash_)
        count += 1
        if count > 0 and (count % 10) == 0:
            sleep(args.sleep_time)

    return 0


if __name__ == '__main__':
    sys.exit(main())



"""Module for Xirvik-specific tasks."""
from setuptools import setup

setup(
    name='xirvik-tools',
    version='0.3.0',
    author='Fa An',
    author_email='2998784916@qq.com',
    packages=['xirvik', 'xirvik.commands', 'xirvik.test'],
    url='https://github.com/Tatsh/xirvik-tools',
    license='LICENSE.txt',
    description='Xirvik (ruTorrent mostly) tools.',
    long_description=open('README.md').read(),
    install_requires=[
        'argcomplete>=1.10.3',
        'benc>=2019.8.1',
        'cached-property>=1.0.0',
        'humanize>=0.5.1',
        'lockfile>=0.10.2',
        'paramiko>=1.17.0',
        'requests>=2.6.0',
        'requests-futures>=1.0.0',
        'six>=1.10.0',
        'typing-extensions>=3.7.4.1',
        'Unidecode>=0.4.19',
    ],
    entry_points={
        'console_scripts': [
            'xirvik-add-ftp-user = xirvik.commands:add_ftp_user',
            'xirvik-auth-ip = xirvik.commands:authorize_ip',
            'xirvik-delete-ftp-user = xirvik.commands:delete_ftp_user',
            'xirvik-delete-old = xirvik.commands.delete_old:main',
            'xirvik-fix-rtorrent = xirvik.commands:fix_rtorrent',
            'xirvik-mirror = xirvik.commands:mirror_main',
            'xirvik-move-by-label = xirvik.commands.move_by_label:main',
            # 'xirvik-move-erroneous = xirvik.commands.move_erroneous:main',
            'xirvik-start-torrents = xirvik.commands:start_torrents',
        ]
    },
    test_suite='xirvik.test',
    tests_require=['coveralls', 'nose', 'requests-mock'])



#!/usr/bin/env python
"""
Deletes old torrents based on specified criteria.
"""
from datetime import datetime, timedelta
from time import sleep
from typing import Any, Callable, Dict, Optional, Tuple, Union
import logging
import sys
import xmlrpc.client as xmlrpc

from requests.exceptions import HTTPError
import argcomplete

from ..client import ruTorrentClient
from .util import common_parser, setup_logging_stdout

TestCallable = Callable[[Dict[str, Any], logging.Logger], Tuple[str, bool]]
TestsDict = Dict[str, Tuple[bool, TestCallable]]


def _test_date_cb(days: int = 14) -> TestCallable:
    def test_date(info: Dict[str, datetime],
                  log: logging.Logger) -> Tuple[str, bool]:
        cond1 = info.get('creation_date')
        cond2 = info.get('state_changed')
        expect = datetime.now() - timedelta(days=days)
        log = setup_logging_stdout()
        log.debug('creation date: %s', cond1)
        log.debug('state changed: %s', cond2)
        log.debug('%s <= %s or', cond1, expect)
        log.debug('    %s <= %s', cond2, expect)
        return (
            'over 14 days seeded',
            bool((cond1 and cond1 <= expect) or (cond2 and cond2 <= expect)),
        )

    return test_date


def _test_ratio(info: Dict[str, float],
                log: logging.Logger) -> Tuple[str, bool]:
    log.debug('ratio: %.2f', info.get('ratio', 0.0))
    return 'ratio >= 1', info.get('ratio', 0.0) >= 1


def _test_ignore(_info: Any) -> Tuple[str, bool]:
    return 'ignoring criteria', True


def main() -> int:
    """Entry point."""
    parser = common_parser()
    parser.add_argument('-a', '--ignore-ratio', action='store_true')
    parser.add_argument('-D', '--ignore-date', action='store_true')
    parser.add_argument('-y', '--dry-run', action='store_true')
    parser.add_argument('--max-attempts', type=int, default=3)
    parser.add_argument('--label')
    parser.add_argument('--sleep-time', type=int, default=10)
    parser.add_argument('--days', type=int, default=14)
    argcomplete.autocomplete(parser)
    args = parser.parse_args()
    log = setup_logging_stdout(verbose=args.verbose)
    client = ruTorrentClient(args.host[0],
                             name=args.username,
                             password=args.password,
                             max_retries=args.max_retries,
                             netrc_path=args.netrc)
    try:
        torrents = client.list_torrents_dict().items()
    except HTTPError:
        log.error('Connection failed on list_torrents() call', file=sys.stderr)
        return 1
    tests: TestsDict = dict(
        ratio=(args.ignore_ratio, _test_ratio),
        date=(args.ignore_date, _test_date_cb(args.days)),
    )
    info: Dict[str, Union[int, str, bool]]
    for hash_, info in torrents:
        if info['left_bytes'] != 0 or info['custom1'] != args.label:
            continue
        reason: Optional[str] = None
        can_delete: bool = False
        for key, (can_ignore, test) in tests.items():
            if can_ignore:
                can_delete = True
                reason = f'ignoring {key}'
                break
            reason, can_delete = test(info, log)
            if can_delete:
                break
        if not can_delete:
            log.info('Cannot delete %s', info['name'])
            continue
        if args.dry_run:
            log.info('Would delete %s, reason: %s', info['name'], reason)
            continue
        else:
            log.info('Deleting %s, reason: %s', info['name'], reason)
        attempts = 0
        while attempts < args.max_attempts:
            attempts += 1
            try:
                client.delete(hash_)
            except xmlrpc.Fault as e:
                log.exception(e)
                sleep_time = args.backoff_factor * (2**(attempts - 1))
                sleep(sleep_time)
            else:
                sleep(args.sleep_time)
                break
    return 0


if __name__ == '__main__':
    sys.exit(main())



"""Logging utility module."""
from functools import lru_cache
from logging.handlers import SysLogHandler
import logging
import sys


@lru_cache()
def get_logger(name: str,
               level: int = logging.INFO,
               verbose: bool = False,
               debug: bool = False,
               syslog: bool = False) -> logging.Logger:
    """
    Set up a logger for sys.stderr/stdout and/or syslog.

    Return a logger object.
    """
    log = logging.getLogger(name)
    if verbose or debug:
        log.setLevel(level if not debug else logging.DEBUG)
        channel = logging.StreamHandler(sys.stdout if debug else sys.stderr)
        channel.setFormatter(logging.Formatter('%(levelname)s - %(message)s'))
        channel.setLevel(level if not debug else logging.DEBUG)
        log.addHandler(channel)
    if syslog:
        log.setLevel(level)
        syslogh = SysLogHandler(address='/dev/log')
        syslogh.setFormatter(logging.Formatter('%(message)s'))
        syslogh.setLevel(logging.INFO)
        log.addHandler(syslogh)
    return log



"""General utility module."""
from hashlib import sha1
from hmac import compare_digest
from os import R_OK, access, environ, stat
from os.path import isdir, join as path_join, realpath
from typing import (Any, BinaryIO, Iterator, List, NoReturn, Optional,
                    Sequence, TypeVar, Union, cast)
import argparse
import platform
import struct
import sys

from benc import decode as bdecode

__all__ = (
    'cleanup_and_exit',
    'ctrl_c_handler',
    'VerificationError',
    'verify_torrent_contents',
    'ReadableDirectoryListAction',
)

T = TypeVar('T')


class ReadableDirectoryAction(argparse.Action):
    """Checks if a directory argument is a directory and is readable."""
    def __call__(self,
                 parser: argparse.ArgumentParser,
                 namespace: argparse.Namespace,
                 values: Optional[Union[str, Sequence[Any]]],
                 option_string: Optional[str] = None) -> None:
        prospective_dir = values
        if not isdir(cast(str, prospective_dir)):
            raise argparse.ArgumentTypeError('%s is not a valid directory' %
                                             (prospective_dir, ))
        # Since macOS 10.15, the Python binary will need access to this
        # directory and a prompt from TCC must appear for this to work
        # Because of TCC becoming more strict, hecking with access() is not
        # reliable on macOS
        if platform.system() == 'Darwin':
            return
        if access(cast(str, prospective_dir), R_OK):
            setattr(namespace, self.dest, realpath(cast(str, prospective_dir)))
            return
        username = environ['USER']
        raise argparse.ArgumentTypeError(
            f'{prospective_dir} is not a readable directory (checking as user '
            f'{username})')


class ReadableDirectoryListAction(argparse.Action):
    """
    Checks if a list of directories argument is a directory and all are
    readable..
    """
    def __call__(self,
                 parser: argparse.ArgumentParser,
                 namespace: argparse.Namespace,
                 values: Optional[Union[str, Sequence[Any]]],
                 option_string: Optional[str] = None) -> None:
        dirs = []
        kwa = dict(self._get_kwargs())
        parent = ReadableDirectoryAction(**kwa)

        for prospective_dir in cast(Sequence[Any], values):
            ns = argparse.Namespace()
            ns.directory = prospective_dir
            parent(parser, ns, prospective_dir, option_string)
            dirs.append(ns.directory)

        setattr(namespace, self.dest, dirs)


def cleanup_and_exit(status: int = 0) -> NoReturn:
    """Called instead of sys.exit(). status is the integer to exit with."""
    sys.exit(status)


def ctrl_c_handler(signum: int, frame: Any) -> NoReturn:
    """Used as a TERM signal handler. Arguments are ignored."""
    raise SystemExit('Signal raised')


def _chunks(l: Sequence[T], n: int) -> Iterator[Sequence[T]]:
    # Source: http://stackoverflow.com/a/312464/374110
    for i in range(0, len(l), n):
        yield l[i:i + n]


def _get_torrent_pieces(filenames: Sequence[str], basepath: str,
                        piece_length: int) -> Iterator[Optional[bytes]]:
    # Yes this is a generator and should not be used any other way (i.e. do not
    # wrap in list()).
    buf = b''
    p_delta = piece_length

    for name in filenames:
        name = path_join(basepath, name)
        try:
            size = stat(name).st_size
        except OSError:
            yield None

        if size <= piece_length and p_delta > size:
            with open(name, 'rb') as f:
                tmp = f.read()
                p_delta -= len(tmp)

                if p_delta <= 0:
                    p_delta = piece_length

                buf += tmp

            continue

        try:
            with open(name, 'rb') as f:
                while True:
                    tmp = f.read(p_delta)
                    p_delta -= len(tmp)

                    if not tmp:
                        break

                    buf += tmp

                    if len(tmp) < p_delta:
                        break

                    if p_delta <= 0:
                        yield buf
                        buf = b''
                        p_delta = piece_length
        except IOError:
            yield None

    # Very last set of bytes of the last file, and this will be <= piece size
    # If this is not returned, a false positive can be given if the last
    # file's last piece is not valid
    if buf:
        yield buf


class VerificationError(Exception):
    """Raised when an error occurs in verify_torrent_contents()."""


def verify_torrent_contents(torrent_file: Union[str, BinaryIO, bytes],
                            path: str) -> None:
    """
    Verify torrent contents.

    Pass a torrent file path and the path to check.
    """
    orig_path = path

    if hasattr(torrent_file, 'seek') and hasattr(torrent_file, 'read'):
        cast(BinaryIO, torrent_file).seek(0)
        torrent = bdecode(cast(BinaryIO, torrent_file).read())
    else:
        try:
            with open(cast(str, torrent_file), 'rb') as f:
                torrent = bdecode(f.read())
        except (IOError, TypeError, ValueError):
            # ValueError for 'embedded null byte' in Python 3.5
            torrent = bdecode(torrent_file)

    path = path_join(path, torrent[b'info'][b'name'].decode('utf-8'))
    is_a_file = False
    try:
        with open(path, 'rb'):
            is_a_file = True
    except IOError:
        pass

    if not isdir(path) and not is_a_file:
        raise IOError('Path specified for torrent data is invalid')

    piece_length = torrent[b'info'][b'piece length']
    piece_hashes = torrent[b'info'][b'pieces']
    piece_hashes = struct.unpack('<{}B'.format(len(piece_hashes)),
                                 piece_hashes)
    piece_hashes = _chunks(piece_hashes, sha1().digest_size)

    try:
        filenames: Union[Iterator[str], List[str]] = ('/'.join(
            [y.decode('utf-8')
             for y in x[b'path']]) for x in torrent[b'info'][b'files'])
    except KeyError as e:
        if e.args[0] != b'files':
            raise e

        # Single file torrent, never has a directory
        filenames = [path]
        path = orig_path

    pieces = _get_torrent_pieces(list(filenames), path, piece_length)

    for known_hash, piece in zip(piece_hashes, pieces):
        known_hash = bytes(known_hash)

        try:
            file_hash = sha1(cast(bytes, piece)).digest()
        except TypeError:
            raise VerificationError('Unable to get hash for piece')

        if not compare_digest(known_hash, file_hash):
            raise VerificationError('Computed hash does not match torrent '
                                    'file\'s hash')



"""Mirror (copy data from remote to local) helper."""
from base64 import b64encode
from logging.handlers import SysLogHandler
from netrc import netrc
from os import chmod, close as close_fd, listdir, makedirs, remove as rm, utime
from os.path import (basename, dirname, expanduser, isdir, join as path_join,
                     realpath, splitext)
from tempfile import gettempdir, mkstemp
from typing import Any, Optional, cast
import argparse
import hashlib
import json
import logging
import os
import signal
import socket
import subprocess as sp
import sys

from lockfile import LockFile, NotLocked
from paramiko import SFTPClient as OriginalSFTPClient
from requests.exceptions import HTTPError
from unidecode import unidecode
import requests

from xirvik.client import (TORRENT_PATH_INDEX, UnexpectedruTorrentError,
                           ruTorrentClient)
from xirvik.log import get_logger
from xirvik.sftp import SFTPClient
from xirvik.util import (ReadableDirectoryListAction, VerificationError,
                         cleanup_and_exit, ctrl_c_handler,
                         verify_torrent_contents)

LOCK = None


def lock_ctrl_c_handler(signum: int, frame: Any) -> None:
    """TERM signal/^C handler."""
    if LOCK:
        try:  # type: ignore[misc]
            LOCK.release()
        except NotLocked:
            pass
    ctrl_c_handler(signum, frame)
    raise SystemExit('Signal raised')


# pylint: disable=protected-access
def mirror(sftp_client: SFTPClient,
           rclient: ruTorrentClient,
           path: str = '.',
           destroot: str = '.',
           keep_modes: bool = True,
           keep_times: bool = True) -> None:
    """
    Mirror a remote directory to local.

    :param sftp_client: must be a valid `xirvik.sftp.SFTPClient` instance.

    `rclient` must be a valid `ruTorrentClient` instance.

    `path` is the remote directory. destroot must be the location where
    destroot/path will be created (the path must not already exist).

    `keep_modes` and `keep_times` are boolean to ensure permissions and time
    are retained respectively.
    """
    cwd = cast(OriginalSFTPClient, sftp_client).getcwd()
    log = logging.getLogger('xirvik')
    for _path, info in sftp_client.listdir_attr_recurse(path=path):
        if info.st_mode & 0o700 == 0o700:
            continue
        dest_path = path_join(destroot, dirname(_path))
        dest = path_join(dest_path, basename(_path))
        if dest_path not in sftp_client._dircache:
            try:
                makedirs(dest_path)
            except OSError:
                pass
            sftp_client._dircache.append(dest_path)
        if isdir(dest):
            continue
        try:
            current_size: Optional[int] = os.stat(dest).st_size
        except OSError:
            current_size = None
        if current_size is None or current_size != info.st_size:
            session = rclient._session
            uri = '{}/downloads{}{}'.format(rclient.http_prefix, cwd,
                                            _path[1:])
            uri = uri.replace('#', '%23')
            log.info('Downloading %s -> %s', uri, dest)
            r = session.get(uri, stream=True)
            r.raise_for_status()
            try:
                total: Optional[int] = int(
                    cast(str, r.headers.get('content-length')))
                log.info('Content-Length: %d', total)
            except (KeyError, ValueError):
                total = None
            with open(dest, 'wb+') as f:
                downloaded = 0
                for chunk in r.iter_content(chunk_size=4096):
                    f.write(chunk)
                    downloaded += len(chunk)
                    done = int(50 * downloaded / cast(int, total))
                    percent = (float(downloaded) /
                               float(cast(int, total))) * 100
                    args = (
                        '=' * done,
                        ' ' * (50 - done),
                        percent,
                    )
                    sys.stdout.write('\r[{}{}] {:.2f}%'.format(*args))
                    sys.stdout.flush()
            sys.stdout.write('\n')
        else:
            log.info('Skipping already downloaded file %s', dest)

        # Okay to fix existing files even if they are already downloaded
        try:
            if keep_modes:
                chmod(dest, info.st_mode)
            if keep_times:
                utime(dest, (
                    info.st_atime,
                    info.st_mtime,
                ))
        except IOError:
            pass


def mirror_main() -> None:
    """Entry point."""
    signal.signal(signal.SIGINT, lock_ctrl_c_handler)
    parser = argparse.ArgumentParser()
    parser.add_argument('-H', '--host', required=True)
    parser.add_argument('-P', '--port', type=int, default=22)
    parser.add_argument('-c', '--netrc-path', default=expanduser('~/.netrc'))
    parser.add_argument('-r',
                        '--resume',
                        action='store_true',
                        help='Resume incomplete files (experimental)')
    parser.add_argument('-T', '--move-to', required=True)
    parser.add_argument('-L', '--label', default='Seeding')
    parser.add_argument('-d', '--debug', action='store_true')
    parser.add_argument('-v', '--verbose', action='store_true')
    parser.add_argument('-s', '--syslog', action='store_true')
    parser.add_argument('--no-preserve-permissions', action='store_false')
    parser.add_argument('--no-preserve-times', action='store_false')
    parser.add_argument('--max-retries', type=int, default=10)
    parser.add_argument('remote_dir', metavar='REMOTEDIR', nargs=1)
    parser.add_argument('local_dir', metavar='LOCALDIR', nargs=1)
    args = parser.parse_args()
    log = get_logger('xirvik',
                     verbose=args.verbose,
                     debug=args.debug,
                     syslog=args.syslog)
    if args.debug:
        for name in ('requests', ):
            _log = logging.getLogger(name)
            formatter = logging.Formatter('%(asctime)s - %(name)s - '
                                          '%(levelname)s - %(message)s')
            channel = logging.StreamHandler(sys.stderr)
            _log.setLevel(logging.DEBUG)
            channel.setLevel(logging.DEBUG)
            channel.setFormatter(formatter)
            _log.addHandler(channel)
    local_dir: str = realpath(args.local_dir[0])
    user_pass = netrc(args.netrc_path).authenticators(args.host)
    assert user_pass is not None
    user, _, password = user_pass
    sftp_host = 'sftp://{user:s}@{host:s}'.format(
        user=user,
        host=args.host,
    )
    lf_hash = hashlib.sha256(json.dumps(
        args._get_kwargs()).encode('utf-8')).hexdigest()
    lf_path = path_join(gettempdir(), 'xirvik-mirror-{}'.format(lf_hash))
    log.debug('Acquiring lock at %s.lock', lf_path)
    _lock = LockFile(lf_path)
    if _lock.is_locked():
        psax = [
            x for x in sp.check_output(('ps',
                                        'ax'), encoding='utf-8').split('\n')
            if sys.argv[0] in x
        ]
        if len(psax) == 1:
            log.info('Breaking lock')
            _lock.break_lock()
    _lock.acquire()
    log.info('Lock acquired')
    log.debug('Local directory to sync to: %s', local_dir)
    log.debug('Read user and password from netrc file')
    log.debug('SFTP URI: %s', sftp_host)
    client = ruTorrentClient(args.host,
                             user,
                             password,
                             max_retries=args.max_retries)
    assumed_path_prefix = '/torrents/{}'.format(user)
    look_for = '{}/{}/'.format(assumed_path_prefix, args.remote_dir[0])
    move_to = '{}/{}'.format(assumed_path_prefix, args.move_to)
    names = {}
    log.debug('Full completed directory path name: %s', look_for)
    log.debug('Moving finished torrents to: %s', move_to)
    log.info('Getting current torrent information (ruTorrent)')
    try:
        torrents = client.list_torrents()
    except requests.exceptions.ConnectionError as e:
        # Assume no Internet connection at this point
        log.error('Failed to connect: %s', e)
        try:
            _lock.release()
        except NotLocked:
            pass
        cleanup_and_exit(1)
    for hash_, v in torrents.items():
        if not v[TORRENT_PATH_INDEX].startswith(look_for):
            continue
        bn = basename(v[TORRENT_PATH_INDEX])
        names[bn] = (
            hash_,
            v[TORRENT_PATH_INDEX],
        )
        log.info(
            'Completed torrent "%s" found with hash %s',
            bn,
            hash_,
        )
    sftp_client_args = dict(
        hostname=args.host,
        username=user,
        password=password,
        port=args.port,
    )
    try:
        with SFTPClient(**sftp_client_args) as sftp_client:
            log.info('Verifying contents of %s with previous '
                     'response', look_for)
            assert sftp_client.chdir(args.remote_dir[0]) is not None
            for item in sftp_client.listdir_iter(read_aheads=10):
                if item.filename not in names:
                    log.error(
                        'File or directory "%s" not found in previous '
                        'response body', item.filename)
                    continue
                log.debug('Found matching torrent "%s" from ls output',
                          item.filename)
            if not names:
                log.info('Nothing found to mirror')
                _lock.release()
                cleanup_and_exit()
            mirror(sftp_client,
                   client,
                   destroot=local_dir,
                   keep_modes=not args.no_preserve_permissions,
                   keep_times=not args.no_preserve_times)
    except (AssertionError, IndexError) as e:
        if args.debug:
            _lock.release()
            raise e
        else:
            log.error(str(e))
        _lock.release()
        cleanup_and_exit()
    _all = names.items()
    exit_status = 0
    bad = []
    for bn, (hash_, fullpath) in _all:
        # There is a warning that can get raised here by urllib3 if
        # Content-Disposition header's filename field has any
        # non-ASCII characters. It is ignorable as the content still gets
        # downloaded correctly
        log.info('Verifying "%s"', bn)
        r, _ = client.get_torrent(hash_)
        try:
            verify_torrent_contents(r.content, local_dir)
        except VerificationError:
            log.error(
                'Could not verify "%s" contents against piece hashes '
                'in torrent file', bn)
            exit_status = 1
            bad.append(hash_)
    # Move to _seeding directory and set label
    # Unfortunately, there is no method, via the API, to do this one HTTP
    #   request
    for bn, (hash_, fullpath) in _all:
        if hash_ in bad:
            continue
        log.info('Moving "%s" to "%s" directory', bn, move_to)
        try:
            client.move_torrent(hash_, move_to)
        except UnexpectedruTorrentError as e:
            log.error(str(e))
    log.info('Setting label to "%s" for downloaded items', args.label)
    client.set_label_to_hashes(hashes=[
        hash_ for bn, (hash, fullpath) in names.items() if hash not in bad
    ],
                               label=args.label)
    if exit_status != 0:
        log.error('Could not verify torrent checksums')
    _lock.release()
    cleanup_and_exit(exit_status)


# pylint: enable=protected-access


def start_torrents() -> None:
    """Uploads torrent files to the server."""
    signal.signal(signal.SIGINT, ctrl_c_handler)

    cache_dir = realpath(expanduser('~/.cache/xirvik'))

    if not isdir(cache_dir):
        makedirs(cache_dir)

    log = logging.getLogger('xirvik-start-torrents')
    parser = argparse.ArgumentParser()

    parser.add_argument('-d', '--debug', action='store_true')
    parser.add_argument('-v', '--verbose', action='store_true')
    parser.add_argument('-c', '--netrc-path', default=expanduser('~/.netrc'))
    # parser.add_argument('-C', '--client', default='rutorrent3')
    parser.add_argument('-H', '--host', nargs=1, required=True)
    parser.add_argument('-p', '--port', nargs=1, default=[443])
    parser.add_argument('--start-stopped', action='store_true')
    parser.add_argument('-s', '--syslog', action='store_true')
    parser.add_argument('directory',
                        metavar='DIRECTORY',
                        action=ReadableDirectoryListAction,
                        nargs='*')

    args = parser.parse_args()
    verbose = args.debug or args.verbose
    log.setLevel(logging.INFO)

    if verbose:
        channel = logging.StreamHandler(
            sys.stdout if args.verbose else sys.stderr)

        channel.setFormatter(logging.Formatter('%(asctime)s - %(message)s'))
        channel.setLevel(logging.INFO if args.verbose else logging.DEBUG)
        log.addHandler(channel)

        if args.debug:
            log.setLevel(logging.DEBUG)

    if args.syslog:
        try:
            syslogh = SysLogHandler(address='/dev/log')
        except (OSError, socket.error):
            syslogh = SysLogHandler(address='/var/run/syslog',
                                    facility=SysLogHandler.LOG_USER)
            syslogh.ident = 'xirvik-start-torrents'  # type: ignore[attr-defined]
            logging.INFO = logging.WARNING

        syslogh.setFormatter(
            logging.Formatter('%(name)s[%(process)d]: %(message)s'))
        syslogh.setLevel(logging.DEBUG if args.debug else logging.INFO)
        log.addHandler(syslogh)

    user_pass = netrc(args.netrc_path).authenticators(args.host[0])
    if not user_pass:
        print((f'Cannot find host {args.host[0]} in netrc. Specify user name '
               'and password'),
              file=sys.stderr)
        sys.exit(1)
    _user, _, _password = user_pass
    post_url = ('https://{host:s}:{port:d}/rtorrent/php/'
                'addtorrent.php?'.format(host=args.host[0], port=args.port[0]))
    form_data = {}
    # rtorrent2/3 params
    # dir_edit - ?
    # tadd_label - Label for the torrents, more param: label=
    # torrent_file - Torrent file blob data
    if args.start_stopped:
        form_data['torrents_start_stopped'] = 'on'
    for d in args.directory:
        for item in listdir(d):
            if not item.lower().endswith('.torrent'):
                continue
            item = path_join(d, item)
            # Workaround for surrogates not allowed error, rename the file
            prefix = '{n:s}-'.format(n=splitext(basename(item))[0])
            fd, name = mkstemp(dir=cache_dir, prefix=prefix, suffix='.torrent')
            close_fd(fd)
            with open(name, 'wb') as w:
                with open(item, 'rb') as r:
                    w.write(r.read())
            old = item
            item = name
            with open(item, 'rb') as torrent_file:
                # Because the server does not understand filename*=UTF8 syntax
                # https://github.com/kennethreitz/requests/issues/2117
                # This is not a huge concern, as the API's "Get .torrent" does
                # not return the file with its original name either
                filename = unidecode(torrent_file.name)
                files = dict(torrent_file=(
                    filename,
                    torrent_file,
                ))
                try:
                    log.info('Uploading torrent %s (actual name: "%s")',
                             basename(item), basename(filename))
                except OSError:
                    pass
                resp = requests.post(post_url, data=form_data, files=files)
                try:
                    resp.raise_for_status()
                except HTTPError as e:
                    log.error('Caught exception: %s', e)
                # Delete original only after successful upload
                log.debug('Deleting %s', old)
                rm(old)


def add_ftp_user() -> int:
    """Adds an FTP user."""
    log = logging.getLogger('xirvik')
    parser = argparse.ArgumentParser()
    parser.add_argument('-u', '--username', required=True)
    parser.add_argument('-P', '--password', required=True)
    parser.add_argument('-r', '--root-directory', default='/')
    parser.add_argument('-d', '--debug', action='store_true')
    parser.add_argument('-v', '--verbose', action='store_true')
    parser.add_argument('-H', '--host', required=True)
    parser.add_argument('-p', '--port', type=int, default=443)
    args = parser.parse_args()
    verbose = args.debug or args.verbose
    log.setLevel(logging.INFO)
    if verbose:
        channel = logging.StreamHandler(
            sys.stdout if args.verbose else sys.stderr)
        channel.setFormatter(logging.Formatter('%(asctime)s - %(message)s'))
        channel.setLevel(logging.INFO if args.verbose else logging.DEBUG)
        log.addHandler(channel)
        if args.debug:
            log.setLevel(logging.DEBUG)
    uri = (f'https://{args.host}:{args.port:d}/userpanel/index.php/'
           'ftp_users/add_user')
    rootdir = args.root_directory if args.root_directory.startswith(
        '/') else f'/{args.root_directory}'
    # Setting read_only=yes does not appear to work
    r = requests.post(uri,
                      data=dict(username=args.username,
                                password_1=args.password,
                                root_folder=rootdir,
                                read_only='no'))
    try:
        r.raise_for_status()
    except HTTPError as e:
        log.exception(e)
        return 1
    return 0


def delete_ftp_user() -> int:
    """Deletes an FTP user."""
    log = logging.getLogger('xirvik')
    parser = argparse.ArgumentParser()
    parser.add_argument('-u', '--username', required=True)
    parser.add_argument('-d', '--debug', action='store_true')
    parser.add_argument('-v', '--verbose', action='store_true')
    parser.add_argument('-H', '--host', required=True)
    parser.add_argument('-p', '--port', type=int, default=443)
    args = parser.parse_args()
    verbose = args.debug or args.verbose
    log.setLevel(logging.INFO)
    if verbose:
        channel = logging.StreamHandler(
            sys.stdout if args.verbose else sys.stderr)
        channel.setFormatter(logging.Formatter('%(asctime)s - %(message)s'))
        channel.setLevel(logging.INFO if args.verbose else logging.DEBUG)
        log.addHandler(channel)
        if args.debug:
            log.setLevel(logging.DEBUG)
    user = b64encode(args.username.encode('utf-8')).decode('utf-8')
    uri = (f'https://{args.host}:{args.port:d}/userpanel/index.php/ftp_users/'
           f'delete/{user}')
    r = requests.get(uri)
    try:
        r.raise_for_status()
    except HTTPError as e:
        log.exception(e)
        return 1
    return 0


def authorize_ip() -> int:
    """Authorises an IP for access to the VM via SSH, removing the previous."""
    log = logging.getLogger('xirvik')
    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--debug', action='store_true')
    parser.add_argument('-v', '--verbose', action='store_true')
    parser.add_argument('-H', '--host', required=True)
    parser.add_argument('-p', '--port', type=int, default=443)
    args = parser.parse_args()
    verbose = args.debug or args.verbose
    log.setLevel(logging.INFO)
    if verbose:
        channel = logging.StreamHandler(
            sys.stdout if args.verbose else sys.stderr)
        channel.setFormatter(logging.Formatter('%(asctime)s - %(message)s'))
        channel.setLevel(logging.INFO if args.verbose else logging.DEBUG)
        log.addHandler(channel)
        if args.debug:
            log.setLevel(logging.DEBUG)
    uri = (f'https://{args.host}:{args.port:d}/userpanel/index.php/'
           'virtual_machine/authorize_ip')
    r = requests.get(uri)
    try:
        r.raise_for_status()
    except HTTPError as e:
        log.exception(e)
        return 1
    return 0


def fix_rtorrent() -> int:
    """
    Restarts the rtorrent service in case ruTorrent cannot connect to it. Not
    guaranteed to fix anything!
    """
    log = logging.getLogger('xirvik')
    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--debug', action='store_true')
    parser.add_argument('-v', '--verbose', action='store_true')
    parser.add_argument('-H', '--host', required=True)
    parser.add_argument('-p', '--port', type=int, default=443)
    args = parser.parse_args()
    verbose = args.debug or args.verbose
    log.setLevel(logging.INFO)
    if verbose:
        channel = logging.StreamHandler(
            sys.stdout if args.verbose else sys.stderr)
        channel.setFormatter(logging.Formatter('%(asctime)s - %(message)s'))
        channel.setLevel(logging.INFO if args.verbose else logging.DEBUG)
        log.addHandler(channel)
        if args.debug:
            log.setLevel(logging.DEBUG)
    log.warning('No guarantees this will work!')
    uri = (f'https://{args.host}:{args.port:d}/userpanel/index.php/services/'
           'restart/rtorrent')
    r = requests.get(uri)
    try:
        r.raise_for_status()
    except HTTPError as e:
        log.exception(e)
        return 1
    return 0
